#property strict
#property version   "0.1"

#define MAKEWORD(low, high) ushort( ((low) & 0xff) | (ushort(high) << 8) )
#ifdef X64
  #define SOCKET          ulong
  #define UINTPTR         ulong
  #define size_t          ulong
#else
  #define SOCKET          uint
  #define UINTPTR         uint
  #define size_t          uint
#endif
#define INVALID_SOCKET    (SOCKET)(~0)
#define WSA_DATA_SZ       420

struct sockaddr_in
{
  short sin_family;
  ushort sin_port;
  uint sin_addr;
  char sin_zero[8];
};

#import "ws2_32.dll"
  int WSAStartup(ushort wVersionRequested, uchar &lpWSAData[]);
  int WSACleanup(void);
  SOCKET socket(int af, int type, int protocol);
  uint inet_addr(char &cp[]);
  ushort htons(ushort hostshort);
  int connect(SOCKET s, sockaddr_in &name, int namelen);
  int closesocket(SOCKET s);
  int recv(SOCKET s, uchar &buf[], int len, int flags);
  int recv(SOCKET s, uchar &buf, int len, int flags);
  int recv(SOCKET s, int &buf, int len, int flags);
  int send(SOCKET s, const uchar &buf[], int len, int flags);
  int send(SOCKET s, const uchar &buf, int len, int flags);
  int send(SOCKET s, const int &buf, int len, int flags);
#import

//+------------------------------------------------------------------+
//| class TCP_Client (blocking read)                                 |
//+------------------------------------------------------------------+
class TCP_Client
{
  SOCKET socket;
  uchar state;
  TCP_Client(const TCP_Client&);
  TCP_Client operator=(TCP_Client &);
public:
  // Address in the "IPv4:port" format.
  TCP_Client(string server_address);
  ~TCP_Client();
  // Return true if the client is connected.
  bool good()const;
  // Receives a data block.
  // At the beginning of each block there must be 4 bytes of int, value
  // of which is equal to block size. After size there is sizeof(prefix)
  // which identifies type of received data in buf. Returns size of 
  // received block (prefix + buf) or 0.
  int recv(uchar &prefix, uchar &buf[]);
  // Sends a block, which is constituted from
  // "prefix + buf[0 : (cnt - 1)]", if cnt > ArraySize(buf) then
  // all array's ellements are included in the block. Function adds
  // 4 byte title which equal to block size . Each buf type
  // requires existence of appropriate send() in #import section. 
  // Returns byte size of sent block or 0.
  template <typename _T>
  int send(uchar prefix, _T &buf[], int cnt = INT_MAX);
};

TCP_Client::TCP_Client(string server_address): socket(INVALID_SOCKET), state(0)
{
  uchar wsa_buf[WSA_DATA_SZ];
  if( WSAStartup(MAKEWORD(2,2), wsa_buf) )
    return;
  this.state |= B'1';
  
  if( (this.socket = socket(2, 1, 0)) == INVALID_SOCKET)
    return;
  this.state |= B'10';
  
  if(true)
  { 
    string server_address_ar[];
    if(StringSplit(server_address, ':', server_address_ar) != 2)
      return;
    char ch_addr[];
    StringToCharArray(server_address_ar[0], ch_addr);
    sockaddr_in addr = {0};
    addr.sin_family = 2;
    addr.sin_addr = inet_addr(ch_addr);
    addr.sin_port = htons( (ushort)StringToInteger(server_address_ar[1]) );
    if( connect(this.socket, addr, sizeof(addr)) )
      return;
    this.state |= B'100';
  }
  
  PRINT("Client is connected");
}

TCP_Client::~TCP_Client()
{
  if( bool(this.state & B'10') )
    closesocket(this.socket);
  if( bool(this.state & B'1') )
    WSACleanup();
}

bool TCP_Client::good()const  {return this.state == B'111';}

int TCP_Client::recv(uchar &prefix, uchar &buf[])
{
  int sz = 0;
  if(true)
  {
    int read = recv(this.socket, sz, sizeof(sz), 8);
    if(read != sizeof(sz))
    {
      this.state &= B'11';
      return 0;
    }
    sz -= sizeof(prefix);
  }
  
  if(true)
  {
    int read = recv(this.socket, prefix, sizeof(prefix), 8);
    if(read != sizeof(prefix))
    {
      this.state &= B'11';
      return 0;
    }
  }
  
  if(sz > 0)
  {
    if( ArraySize(buf) < sz )
      if(ArrayResize(buf, sz) < sz)
        return 0;
    int read = recv(this.socket, buf, sz, 8);
    if(read != sz)
    {
      this.state &= B'11';
      return 0;
    }
  }
  return sz + sizeof(prefix);
}

template <typename _T>
int TCP_Client::send(uchar prefix, _T &buf[], int cnt)
{
  int size = sizeof(prefix) + (cnt > ArraySize(buf) ?
                                ArraySize(buf) * sizeof(buf[0]) : 
                                cnt * sizeof(buf[0]));
    
  if(true)
  {
    int sent = send(this.socket, size, sizeof(size), 0);
    if( sent != sizeof(size) )
    {
      this.state &= B'11';
      return 0;
    }
  }
  
  if(true)
  {
    int sent = send(this.socket, prefix, sizeof(prefix), 0);
    if( sent != sizeof(prefix) )
    {
      this.state &= B'11';
      return 0;
    }
  }
  
  if(size - sizeof(prefix) > 0)
  {
    int sent = send(this.socket, buf, size - sizeof(prefix), 0);
    if( sent != size - sizeof(prefix) )
    {
      this.state &= B'11';
      return 0;
    }
  }
  return size;
}